#include "driver.h"
#include "utility.h"
#include "ioctl.h"
#include "usb.h"
#include "usbgetdescriptor.h"
#include "usbrequests.h"
#include "usbezusb.h"
#include "dispatchioctl.h"

NTSTATUS
DispatchDeviceControl(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
    )
/*
Routine Description:
   Hier worden alle DeviceIoControl codes afgehandeld.
   Er kan code worden toegevoegd om specifieke IOCTL codes af te handelen.
Arguments:
   DeviceObject - pointer naar het Device Object voor deze instantie van het USB Device
Return Value:
   NT status code
*/
{
   PIO_STACK_LOCATION irpStack;
   PVOID ioBuffer;
   ULONG inputBufferLength;
   ULONG outputBufferLength;
   PDEVICE_EXTENSION pDeviceExtension = (PDEVICE_EXTENSION)DeviceObject->DeviceExtension;
   ULONG ioControlCode;
   NTSTATUS ntStatus;
   ULONG length;
   KdPrint((debugPrefix "DispatchDeviceControl\n"));

   if (!UtilLockDeviceObject(DeviceObject))
      return UtilCompleteRequest(Irp, STATUS_DELETE_PENDING, 0);

   // verkrijg een pointer naar de huidige stack locatie in de IRP.
   // dat is de plek waar de functiecodes en parameters zich bevinden.
   irpStack = IoGetCurrentIrpStackLocation(Irp);

   Irp->IoStatus.Status = STATUS_SUCCESS;
   Irp->IoStatus.Information = 0;

   ioBuffer           = Irp->AssociatedIrp.SystemBuffer;
   inputBufferLength  = irpStack->Parameters.DeviceIoControl.InputBufferLength;
   outputBufferLength = irpStack->Parameters.DeviceIoControl.OutputBufferLength;

   ioControlCode = irpStack->Parameters.DeviceIoControl.IoControlCode;

   // handel user mode I/O Control Codes af
   switch (ioControlCode)
   {
      case IOCTL_Ezusb_VENDOR_REQUEST: // nodig voor het resetten van de 8051 processor
      {
         if (length = UsbVendorRequest(DeviceObject, (PVENDOR_REQUEST_IN)ioBuffer))
         {
            Irp->IoStatus.Information = length;
            Irp->IoStatus.Status = STATUS_SUCCESS;
         }
         else
            Irp->IoStatus.Status = STATUS_SUCCESS;
         KdPrint((debugPrefix "DispatchDeviceControl: IOCTL_Ezusb_VENDOR_REQUEST (resultaat = %d bytes)\n", length));
         break;
      }
      case IOCTL_EZUSB_ANCHOR_DOWNLOAD: // nodig voor het uploaden van de firmware
      {
         PANCHOR_DOWNLOAD_CONTROL downloadControl = (PANCHOR_DOWNLOAD_CONTROL)ioBuffer;

         // controleer de parameters
         if (inputBufferLength != sizeof(ANCHOR_DOWNLOAD_CONTROL) || outputBufferLength == 0)
         {
            KdPrint((debugPrefix "DispatchDeviceControl: IOCTL_EZUSB_ANCHOR_DOWNLOAD: ongeldige parameter\n"));
            ntStatus = STATUS_INVALID_PARAMETER;
         }
         else
         {
            KdPrint((debugPrefix "DispatchDeviceControl: IOCTL_EZUSB_ANCHOR_DOWNLOAD\n"));
            ntStatus = EzusbAnchorDownload(DeviceObject,
                                         downloadControl->Offset,
                                         (PUCHAR)MmGetSystemAddressForMdl(Irp->MdlAddress),
                                         outputBufferLength);
         }
         break;
      }
      case IOCTL_Ezusb_RESETPIPE:
      {
         ULONG pipenum = *((PULONG)ioBuffer);
         KdPrint((debugPrefix "DispatchDeviceControl: IOCTL_Ezusb_RESETPIPE\n"));
         Irp->IoStatus.Status = UsbResetPipe(DeviceObject, pipenum);
         break;
      }
      case IOCTL_Ezusb_ABORTPIPE:
      {
         int pipenum = *((PULONG)ioBuffer);
         KdPrint((debugPrefix "DispatchDeviceControl: IOCTL_Ezusb_ABORTPIPE\n"));
         UsbAbortPipe(DeviceObject, (USBD_PIPE_HANDLE)pDeviceExtension->Interface->Pipes[pipenum].PipeHandle);
         Irp->IoStatus.Information = 0;
         Irp->IoStatus.Status = STATUS_SUCCESS;
         break;
      }
      case IOCTL_Ezusb_GET_PIPE_INFO:
      {
         // inputs  - geen
         // outputs - de interface informatie die in de Device Extension is opgeslagen
         //           wordt gekopieerd naar de uitvoerbuffer welke door de IOS zal
         //           worden zichtbaargemaakt in User Mode
         length = 0;
         KdPrint((debugPrefix "DispatchDeviceControl: IOCTL_Ezusb_GET_PIPE_INFO\n"));
         if (pDeviceExtension->Interface)
         {
            RtlCopyMemory((PUCHAR)ioBuffer + length,
                          (PUCHAR)pDeviceExtension->Interface,
                          pDeviceExtension->Interface->Length);
            length += pDeviceExtension->Interface->Length;
         }
         Irp->IoStatus.Information = length;
         Irp->IoStatus.Status = STATUS_SUCCESS;
         break;
      }
      case IOCTL_Ezusb_GET_DEVICE_DESCRIPTOR:
      {
         // inputs  - pointer naar een buffer voor de Descriptor data
         // outputs - als er Device Descriptor data is teruggegeven door het device wordt
         //           deze in de systeembuffer gezet en wordt het "length" veld in het
         //           "Information" veld van de Irp ingesteld, zodat het systeem de buffer
         //           terugkopieert naar de User Mode buffer in
         KdPrint((debugPrefix "DispatchDeviceControl: IOCTL_Ezusb_GET_DEVICE_DESCRIPTOR wordt aangeroepen\n"));
         length = UsbGetDeviceDescriptor(DeviceObject, ioBuffer, /*veiligheidslek??*/sizeof(USB_DEVICE_DESCRIPTOR)/*outputBufferLength*/);
         KdPrint((debugPrefix "DispatchDeviceControl: IOCTL_Ezusb_GET_DEVICE_DESCRIPTOR is aangeroepen (teruggegeven lengte = %d bytes)\n", length));
         Irp->IoStatus.Information = length;
         Irp->IoStatus.Status = STATUS_SUCCESS;
         break;
      }
      case IOCTL_Ezusb_GET_STRING_DESCRIPTOR:
      {
         KdPrint((debugPrefix "DispatchDeviceControl: IOCTL_Ezusb_GET_STRING_DESCRIPTOR\n"));
         if (
            (inputBufferLength = sizeof(GET_STRING_DESCRIPTOR_IN))
            &&
            (outputBufferLength > 0)
         )
         {
            PGET_STRING_DESCRIPTOR_IN input = ioBuffer;
            if (length = UsbGetStringDescriptor(DeviceObject,
                                                ioBuffer,
                                                outputBufferLength,
                                                input->Index,
                                                input->LanguageId))
            {
               Irp->IoStatus.Information = length;
               Irp->IoStatus.Status = STATUS_SUCCESS;
            }
            else
               Irp->IoStatus.Status = STATUS_UNSUCCESSFUL;
         }
         else
            Irp->IoStatus.Status = STATUS_INVALID_PARAMETER;
         break;
      }
	  case IOCTL_Ezusb_GET_CONFIGURATION_DESCRIPTOR:
      {
		 // inputs  - pointer naar een buffer waarin de Descriptor data geplaatst wordt
		 // outputs - als er Configuration Descriptor data wordt teruggegeven door het apparaat,
         //           wordt deze in de systeembuffer geplaatst en vervolgens wordt de lengte
		 //           in het Information veld van de Irp geplaatst, wat ervoor zal zorgen dat
		 //           het systeem de buffer terugkopieert naar de User Mode buffer
		 length = UsbGetDescriptor(DeviceObject, ioBuffer, outputBufferLength, USB_CONFIGURATION_DESCRIPTOR_TYPE);
		 KdPrint((debugPrefix "DispatchDeviceControl: IOCTL_Ezusb_GET_CONFIGURATION_DESCRIPTOR (teruggegeven lengte = %d bytes)\n"));
		 Irp->IoStatus.Information = length;
		 Irp->IoStatus.Status = STATUS_SUCCESS;
		 break;
      }
      case IOCTL_Ezusb_SETINTERFACE:
      {
         PSET_INTERFACE_IN input = ioBuffer;
		 KdPrint((debugPrefix "DispatchDeviceControl: IOCTL_Ezusb_SETINTERFACE\n"));
         Irp->IoStatus.Status = UsbSetInterface(DeviceObject,
                                                input->interfaceNum,
                                                input->alternateSetting);
         Irp->IoStatus.Status = 0;
         break;
      }
      case IOCTL_Ezusb_RESET:
      {
		 KdPrint((debugPrefix "DispatchDeviceControl: IOCTL_Ezusb_RESET\n"));
         UsbResetParentPort(DeviceObject);
         break;
      }
      case IOCTL_EZUSB_BULK_WRITE: // nodig voor bulk overdracht
      case IOCTL_EZUSB_BULK_READ: // nodig voor bulk overdracht
      {
		 KdPrint((debugPrefix "DispatchDeviceControl: IOCTL_EZUSB_BULK_WRITE/IOCTL_EZUSB_BULK_READ\n"));
         Irp->IoStatus.Status = UsbTransfer(DeviceObject, Irp);
         break;
      }
      case IOCTL_EZUSB_VENDOR_OR_CLASS_REQUEST:
      {
		 KdPrint((debugPrefix "DispatchDeviceControl: IOCTL_EZUSB_VENDOR_OR_CLASS_REQUEST\n"));
         Irp->IoStatus.Status = UsbVendorRequestMDL(DeviceObject, Irp);
         break;
      }
      case IOCTL_EZUSB_GET_LAST_ERROR:
      {
         // controleer dat de uitvoerbuffer goed is en kopieer dan de meest recente URB status
         // vanuit de Device Extension daarnaartoe
		 KdPrint((debugPrefix "DispatchDeviceControl: IOCTL_EZUSB_GET_LAST_ERROR\n"));
         if (outputBufferLength >= sizeof(ULONG))
         {
            *((PULONG)ioBuffer) = pDeviceExtension->LastFailedUrbStatus;
            Irp->IoStatus.Status = STATUS_SUCCESS;
            Irp->IoStatus.Information = sizeof(ULONG);
         }
         else
            Irp->IoStatus.Status = STATUS_UNSUCCESSFUL;
         break;
      }
      case IOCTL_EZUSB_GET_DRIVER_VERSION:
      {
		 KdPrint((debugPrefix "DispatchDeviceControl: IOCTL_EZUSB_GET_DRIVER_VERSION\n"));
         if (outputBufferLength >= sizeof(EZUSB_DRIVER_VERSION))
         {
            PEZUSB_DRIVER_VERSION version = (PEZUSB_DRIVER_VERSION)ioBuffer;
            version->MajorVersion = EZUSB_MAJOR_VERSION;
            version->MinorVersion = EZUSB_MINOR_VERSION;
            version->BuildVersion = EZUSB_BUILD_VERSION;

            Irp->IoStatus.Status = STATUS_SUCCESS;
            Irp->IoStatus.Information = sizeof(EZUSB_DRIVER_VERSION);
         }
         else
            Irp->IoStatus.Status = STATUS_UNSUCCESSFUL;
         break;
      }
      case IOCTL_EZUSB_SET_FEATURE:
      {
		 KdPrint((debugPrefix "DispatchDeviceControl: IOCTL_EZUSB_SET_FEATURE\n"));
         // controleren van invoerparameters
         if (inputBufferLength != sizeof(SET_FEATURE_CONTROL))
         {
            Irp->IoStatus.Information = 0; 
            Irp->IoStatus.Status = STATUS_INVALID_PARAMETER;
            break;
         }
         UsbSetFeature(DeviceObject, (PSET_FEATURE_CONTROL)Irp->AssociatedIrp.SystemBuffer);
         Irp->IoStatus.Status = STATUS_UNSUCCESSFUL;
         break;
      }
      default:
         Irp->IoStatus.Status = STATUS_INVALID_PARAMETER;
   }
   UtilUnlockDeviceObject(DeviceObject);
   ntStatus = Irp->IoStatus.Status;
   IoCompleteRequest(Irp, IO_NO_INCREMENT);
   return ntStatus;
}
