#include "driver.h"
#include "utility.h"
#include "usb.h"
#include "usbrequests.h"
#include "dispatchpnp.h"

NTSTATUS
UtilForwardAndWait(
   IN PDEVICE_OBJECT DeviceObject,
   IN PIRP Irp
   )
/*
Routine Description:
   Geef het verzoek door aan een lagere driver en wacht op completering

   Het enige doel van deze routine in deze driver is het naar beneden doorgeven van
   IRP_MN_START_DEVICE verzoeken en wachten op het PDO om ze af te handelen.
   
   De processor moet zich bevinden op IRL = PASSIVE_LEVEL omdat deze functie
   een kernel event object initialiseert en daarop wacht gedurende een tijdsduur ongelijk aan nul.
Arguments:
   DeviceObject - pointer naar een Device Object
   Irp          - pointer naar een I/O Request Packet
Return Value:
   STATUS_SUCCESS indien succesvol,
   anders STATUS_UNSUCCESSFUL
*/
{
   KEVENT event;
   PDEVICE_EXTENSION pDeviceExtension = (PDEVICE_EXTENSION)DeviceObject->DeviceExtension;
   NTSTATUS ntStatus;

   ASSERT(KeGetCurrentIrql() == PASSIVE_LEVEL);
	
   // initialiseer een kernel event object om te gebruiken bij het wachten tot de lagereliggende driver
   // klaar is met het verwerken van het object.
   KeInitializeEvent(&event, NotificationEvent, FALSE);

   IoCopyCurrentIrpStackLocationToNext(Irp);
   IoSetCompletionRoutine(Irp,
                          (PIO_COMPLETION_ROUTINE)IoCompletion,
                          (PVOID)&event,
                          TRUE,
                          TRUE,
                          TRUE);

   if ((ntStatus = IoCallDriver(pDeviceExtension->StackDeviceObject, Irp)) == STATUS_PENDING)
   {
      KeWaitForSingleObject(&event, Executive, KernelMode, FALSE, NULL);
      ntStatus = Irp->IoStatus.Status;
   }

   return ntStatus;
}

NTSTATUS
PnpDefault(
   IN PDEVICE_OBJECT DeviceObject,
   IN PIRP Irp
   )
{
   PDEVICE_EXTENSION pDeviceExtension = (PDEVICE_EXTENSION)DeviceObject->DeviceExtension;
   IoSkipCurrentIrpStackLocation(Irp);
   return IoCallDriver(pDeviceExtension->StackDeviceObject, Irp);
}

NTSTATUS
PnpStartDevice(
   IN PDEVICE_OBJECT DeviceObject,
   IN PIRP Irp
   )
{
   PDEVICE_EXTENSION pDeviceExtension = (PDEVICE_EXTENSION)DeviceObject->DeviceExtension;
   NTSTATUS ntStatus;
   KdPrint((debugPrefix "PnpStartDevice: begin\n"));

   // geef eerst alle lagerliggende drivers de mogelijkheid om het verzoek af te handelen
   ntStatus = UtilForwardAndWait(DeviceObject, Irp);
   if (!NT_SUCCESS(ntStatus))
   {
      KdPrint((debugPrefix "PnpStartDevice: mislukt (ntStatus = %x)\n", ntStatus));
      return UtilCompleteRequest(Irp, ntStatus, Irp->IoStatus.Information);
   }

   // zorg dat het device gestart wordt
   if (NT_SUCCESS(ntStatus = UsbStartDevice(DeviceObject)))
      pDeviceExtension->Started = TRUE;

   KdPrint((debugPrefix "PnpStartDevice: einde\n", ntStatus));
   return UtilCompleteRequest(Irp, ntStatus, 0);
}

NTSTATUS
RemoveDeviceObject(
    IN  PDEVICE_OBJECT DeviceObject
    )
/*
Routine Description:
    Verwijdert een bepaald Device Object en ruimt bijbehorende zaken op.
Arguments:
    DeviceObject - pointer naar het Device Object voor deze instantie van het USB device.
Return Value:
    NT status code
*/
{
   PDEVICE_EXTENSION pDeviceExtension = (PDEVICE_EXTENSION)DeviceObject->DeviceExtension;
   NTSTATUS ntStatus = STATUS_SUCCESS;
   KdPrint((debugPrefix "RemoveDeviceObject: begin\n"));

   if (pDeviceExtension->DeviceDescriptor)
      ExFreePool(pDeviceExtension->DeviceDescriptor);

   // geef interfacestructuren binnen de Device Extension vrij
   if (pDeviceExtension->Interface != NULL)
      ExFreePool(pDeviceExtension->Interface);

   // verwijder de symbolic link van het Device Object
   if (pDeviceExtension->NeedCleanup)
   {
      UNICODE_STRING deviceLinkUnicodeString;

      pDeviceExtension->NeedCleanup = FALSE;

      RtlInitUnicodeString(&deviceLinkUnicodeString, pDeviceExtension->DeviceLinkNameBuffer);
      IoDeleteSymbolicLink(&deviceLinkUnicodeString);
   }

   IoDetachDevice(pDeviceExtension->StackDeviceObject);
   IoDeleteDevice(DeviceObject);

   KdPrint((debugPrefix "RemoveDeviceObject: einde (ntStatus = %x)\n", ntStatus));
   return ntStatus;
}

NTSTATUS
PnpRemoveDevice(
   IN PDEVICE_OBJECT DeviceObject,
   IN PIRP Irp
   )
{
   PDEVICE_EXTENSION pDeviceExtension = (PDEVICE_EXTENSION)DeviceObject->DeviceExtension;
   ULONG i;

   // zet de "removing" vlag aan om nieuwe "I/O's" te vermijden
   pDeviceExtension->removing = TRUE;

   // brute force - verzend een "abort pipe" bericht naar alle pipes om wachtende overdracht
   // te annuleren. dit lost het probleem op dat de driver blokkeert bij een "REMOVE" bericht
   // als er een wachtende overdracht is.
   for (i = 0; i < pDeviceExtension->Interface->NumberOfPipes; i++)
      UsbAbortPipe(DeviceObject, (USBD_PIPE_HANDLE)pDeviceExtension->Interface->Pipes[i].PipeHandle);

   // een keer voor LockDevice aan het begin van dispatch
   UtilUnlockDeviceObject(DeviceObject);
   // een keer voor initialisatie tijdens AddDevice
   UtilUnlockDeviceObject(DeviceObject);
   KeWaitForSingleObject(&pDeviceExtension->evRemove, Executive, KernelMode, FALSE, NULL);

   RemoveDeviceObject(DeviceObject);

   return PnpDefault(DeviceObject, Irp); // IoStatus is al gecompleteerd op een lager niveau
}

NTSTATUS
PnpStopDevice(
   IN PDEVICE_OBJECT DeviceObject,
   IN PIRP Irp
   )
{
   // laat de aanvraag afdalen over de stack
   PnpDefault(DeviceObject, Irp);
   return UsbStopDevice(DeviceObject);
}

NTSTATUS
DispatchPnp(
   IN PDEVICE_OBJECT DeviceObject,
   IN PIRP           Irp
   )
/*
Routine Description:
   Verwerk Plug and Play IRPs die naar dit device verzonden zijn.
Arguments:
   DeviceObject - pointer naar een Device Object
   Irp - pointer naar een I/O Request Packet
Return Value:
   NTSTATUS
*/
{
   PIO_STACK_LOCATION irpStack;
   PDEVICE_EXTENSION pDeviceExtension = DeviceObject->DeviceExtension;
   NTSTATUS ntStatus;
   ULONG minorFunction;
   KdPrint((debugPrefix "DispatchPnp: begin\n"));

   if (!UtilLockDeviceObject(DeviceObject))
      return UtilCompleteRequest(Irp, STATUS_DELETE_PENDING, 0);

   // verkrijg een pointer naar de huidige stack locatie in de IRP.
   // dat is de plek waar de functiecodes en parameters zich bevinden.
   irpStack = IoGetCurrentIrpStackLocation (Irp);

   ASSERT(irpStack->MajorFunction == IRP_MJ_PNP);

   switch (minorFunction = irpStack->MinorFunction)
   {
      case IRP_MN_START_DEVICE:
      {
         KdPrint((debugPrefix "DispatchPnp: IRP_MN_START_DEVICE\n"));
         ntStatus = PnpStartDevice(DeviceObject, Irp);
         break; 
      }
      case IRP_MN_REMOVE_DEVICE:
      {
         KdPrint((debugPrefix "DispatchPnp: IRP_MN_REMOVE_DEVICE\n"));
         ntStatus = PnpRemoveDevice(DeviceObject, Irp);
         break;
      }
      case IRP_MN_STOP_DEVICE:
      {
         KdPrint((debugPrefix "DispatchPnp: IRP_MN_STOP_DEVICE\n"));
         ntStatus = PnpStopDevice(DeviceObject, Irp);
         break;
      }
      case IRP_MN_QUERY_CAPABILITIES:
      {
         // deze code is afkomstig van Walter Oney
         PDEVICE_CAPABILITIES pdc = irpStack->Parameters.DeviceCapabilities.Capabilities;

         KdPrint((debugPrefix "DispatchPnp: IRP_MN_QUERY_CAPABILITIES\n"));

         // controleer om zeker te zijn dat we weten hoe deze versie van de Capabilities structure moet worden verwerkt
	     if (pdc->Version < 1)
         {
		    ntStatus = PnpDefault(DeviceObject, Irp);
            break;
         }

         ntStatus = UtilForwardAndWait(DeviceObject, Irp);
	     if (NT_SUCCESS(ntStatus)) // IRP succesvol
         {
      		pdc = irpStack->Parameters.DeviceCapabilities.Capabilities;
            // het inschakelen van dit veld voorkomt dat NT5 de gebruiker ervoor waarschuwt
            // dat het apparaat verwijderd is.
		    pdc->SurpriseRemovalOK = TRUE;
         }
         ntStatus = UtilCompleteRequest(Irp, ntStatus, Irp->IoStatus.Information);
         break; 
      }

      // alle overige PNP IRP's worden eenvoudigweg naar beneden doorgegeven via de stack met de standaardafhandeling
      case IRP_MN_QUERY_REMOVE_DEVICE:
      {
         KdPrint((debugPrefix "DispatchPnp: onafgehandelde PnP MinorCode IRP_MN_QUERY_REMOVE_DEVICE wordt naar beneden doorgegeven\n"));
         ntStatus = PnpDefault(DeviceObject, Irp);
         break;
      }
      case IRP_MN_CANCEL_REMOVE_DEVICE:
      {
         KdPrint((debugPrefix "DispatchPnp: onafgehandelde PnP MinorCode IRP_MN_CANCEL_REMOVE_DEVICE wordt naar beneden doorgegeven\n"));
         ntStatus = PnpDefault(DeviceObject, Irp);
         break;
      }
      case IRP_MN_QUERY_STOP_DEVICE:
      {
         KdPrint((debugPrefix "DispatchPnp: onafgehandelde PnP MinorCode IRP_MN_QUERY_STOP_DEVICE wordt naar beneden doorgegeven\n"));
         ntStatus = PnpDefault(DeviceObject, Irp);
         break;
      }
      case IRP_MN_CANCEL_STOP_DEVICE:
      {
         KdPrint((debugPrefix "DispatchPnp: onafgehandelde PnP MinorCode IRP_MN_CANCEL_STOP_DEVICE wordt naar beneden doorgegeven\n"));
         ntStatus = PnpDefault(DeviceObject, Irp);
         break;
      }
      case IRP_MN_QUERY_DEVICE_RELATIONS:
      {
         KdPrint((debugPrefix "DispatchPnp: onafgehandelde PnP MinorCode IRP_MN_QUERY_DEVICE_RELATIONS wordt naar beneden doorgegeven\n"));
         ntStatus = PnpDefault(DeviceObject, Irp);
         break;
      }
      case IRP_MN_QUERY_INTERFACE:
      {
         KdPrint((debugPrefix "DispatchPnp: onafgehandelde PnP MinorCode IRP_MN_QUERY_INTERFACE wordt naar beneden doorgegeven\n"));
         ntStatus = PnpDefault(DeviceObject, Irp);
         break;
      }
      case IRP_MN_QUERY_RESOURCES:
      {
         KdPrint((debugPrefix "DispatchPnp: onafgehandelde PnP MinorCode IRP_MN_QUERY_RESOURCES wordt naar beneden doorgegeven\n"));
         ntStatus = PnpDefault(DeviceObject, Irp);
         break;
      }
      case IRP_MN_QUERY_RESOURCE_REQUIREMENTS:
      {
         KdPrint((debugPrefix "DispatchPnp: onafgehandelde PnP MinorCode IRP_MN_QUERY_RESOURCE_REQUIREMENTS wordt naar beneden doorgegeven\n"));
         ntStatus = PnpDefault(DeviceObject, Irp);
         break;
      }
      case IRP_MN_QUERY_DEVICE_TEXT:
      {
         KdPrint((debugPrefix "DispatchPnp: onafgehandelde PnP MinorCode IRP_MN_QUERY_DEVICE_TEXT wordt naar beneden doorgegeven\n"));
         ntStatus = PnpDefault(DeviceObject, Irp);
         break;
      }
      case IRP_MN_FILTER_RESOURCE_REQUIREMENTS:
      {
         KdPrint((debugPrefix "DispatchPnp: onafgehandelde PnP MinorCode IRP_MN_FILTER_RESOURCE_REQUIREMENTS wordt naar beneden doorgegeven\n"));
         ntStatus = PnpDefault(DeviceObject, Irp);
         break;
      }
      case IRP_MN_READ_CONFIG:
      {
         KdPrint((debugPrefix "DispatchPnp: onafgehandelde PnP MinorCode IRP_MN_READ_CONFIG wordt naar beneden doorgegeven\n"));
         ntStatus = PnpDefault(DeviceObject, Irp);
         break;
      }
      case IRP_MN_WRITE_CONFIG:
      {
         KdPrint((debugPrefix "DispatchPnp: onafgehandelde PnP MinorCode IRP_MN_WRITE_CONFIG wordt naar beneden doorgegeven\n"));
         ntStatus = PnpDefault(DeviceObject, Irp);
         break;
      }
      case IRP_MN_EJECT:
      {
         KdPrint((debugPrefix "DispatchPnp: onafgehandelde PnP MinorCode IRP_MN_EJECT wordt naar beneden doorgegeven\n"));
         ntStatus = PnpDefault(DeviceObject, Irp);
         break;
      }
      case IRP_MN_SET_LOCK:
      {
         KdPrint((debugPrefix "DispatchPnp: onafgehandelde PnP MinorCode IRP_MN_SET_LOCK wordt naar beneden doorgegeven\n"));
         ntStatus = PnpDefault(DeviceObject, Irp);
         break;
      }
      case IRP_MN_QUERY_ID:
      {
         KdPrint((debugPrefix "DispatchPnp: onafgehandelde PnP MinorCode IRP_MN_QUERY_ID wordt naar beneden doorgegeven\n"));
         ntStatus = PnpDefault(DeviceObject, Irp);
         break;
      }
      case IRP_MN_QUERY_PNP_DEVICE_STATE:
      {
         KdPrint((debugPrefix "DispatchPnp: onafgehandelde PnP MinorCode IRP_MN_QUERY_PNP_DEVICE_STATE wordt naar beneden doorgegeven\n"));
         ntStatus = PnpDefault(DeviceObject, Irp);
         break;
      }
      case IRP_MN_QUERY_BUS_INFORMATION:
      {
         KdPrint((debugPrefix "DispatchPnp: onafgehandelde PnP MinorCode IRP_MN_QUERY_BUS_INFORMATION wordt naar beneden doorgegeven\n"));
         ntStatus = PnpDefault(DeviceObject, Irp);
         break;
      }
      case IRP_MN_DEVICE_USAGE_NOTIFICATION:
      {
         KdPrint((debugPrefix "DispatchPnp: onafgehandelde PnP MinorCode IRP_MN_DEVICE_USAGE_NOTIFICATION wordt naar beneden doorgegeven\n"));
         ntStatus = PnpDefault(DeviceObject, Irp);
         break;
      }
      case IRP_MN_SURPRISE_REMOVAL:
      {
         KdPrint((debugPrefix "DispatchPnp: onafgehandelde PnP MinorCode IRP_MN_SURPRISE_REMOVAL wordt naar beneden doorgegeven\n"));
         ntStatus = PnpDefault(DeviceObject, Irp);
         break;
      }
      default:
      {
         KdPrint((debugPrefix "DispatchPnp: onafgehandelde PnP MinorCode 0x%x wordt naar beneden doorgegeven\n", minorFunction));
         ntStatus = PnpDefault(DeviceObject, Irp);
      }
   }

   if (minorFunction != IRP_MN_REMOVE_DEVICE)
      UtilUnlockDeviceObject(DeviceObject);

   KdPrint((debugPrefix "DispatchPnp: einde (ntStatus = %x)\n", ntStatus));
   return ntStatus;
} // DispatchPnp
