#include "driver.h"
#include "usbrequests.h"
#include "usb.h"

NTSTATUS
UsbStartDevice(
    IN  PDEVICE_OBJECT DeviceObject
    )
/*
Routine Description:
   Initialiseert een bepaalde instantie van het USB device op de USB bus.
Arguments:
   DeviceObject - pointer naar een Device Object voor deze instantie van het USB device
Return Value:
   NT status code
*/
{
    PDEVICE_EXTENSION pDeviceExtension = (PDEVICE_EXTENSION)DeviceObject->DeviceExtension;
    NTSTATUS ntStatus;
    PUSB_DEVICE_DESCRIPTOR deviceDescriptor = NULL;
    PURB urb; 
    KdPrint((debugPrefix "UsbStartDevice: begin\n"));

    pDeviceExtension->NeedCleanup = TRUE;

    // reserveer wat geheugen uit de niet-gepagineerde pool (vast, gelockt systeemgeheugen)
    // voor gebruik door het USB Request Block (URB) voor het specifieke USB Request dat
    // hieronder wordt uitgevoerd (een USB device request).
    if (urb = ExAllocatePool(NonPagedPool, sizeof(struct _URB_CONTROL_DESCRIPTOR_REQUEST)))
    {
        ULONG size = sizeof(USB_DEVICE_DESCRIPTOR);

        // alloceer wat niet-gepagineerd geheugen voor de inhoud van de Device Descriptor
        if (deviceDescriptor = ExAllocatePool(NonPagedPool, size))
        {
            // gebruik een macro uit de standaard USB header bestanden om de URB op te bouwen
            UsbBuildGetDescriptorRequest(urb,
                                         (USHORT)sizeof(struct _URB_CONTROL_DESCRIPTOR_REQUEST),
                                         USB_DEVICE_DESCRIPTOR_TYPE,
                                         0,
                                         0,
                                         deviceDescriptor,
                                         NULL,
                                         size,
                                         NULL);
            // vraag de Device Descriptor op en print de informatie naar de debugger indien succesvol
            if (NT_SUCCESS(ntStatus = UsbCallUSBD(DeviceObject, urb)))
            {
                KdPrint((debugPrefix "UsbStartDevice: Device Descriptor = %x, lengte = %x\n",
                         deviceDescriptor,
                         urb->UrbControlDescriptorRequest.TransferBufferLength));

                KdPrint((debugPrefix "UsbStartDevice: USB Device Descriptor:\n"));
                KdPrint((debugPrefix "UsbStartDevice: =======================\n"));
                KdPrint((debugPrefix "UsbStartDevice: bLength %d\n", deviceDescriptor->bLength));
                KdPrint((debugPrefix "UsbStartDevice: bDescriptorType 0x%x\n", deviceDescriptor->bDescriptorType));
                KdPrint((debugPrefix "UsbStartDevice: bcdUSB 0x%x\n", deviceDescriptor->bcdUSB));
                KdPrint((debugPrefix "UsbStartDevice: bDeviceClass 0x%x\n", deviceDescriptor->bDeviceClass));
                KdPrint((debugPrefix "UsbStartDevice: bDeviceSubClass 0x%x\n", deviceDescriptor->bDeviceSubClass));
                KdPrint((debugPrefix "UsbStartDevice: bDeviceProtocol 0x%x\n", deviceDescriptor->bDeviceProtocol));
                KdPrint((debugPrefix "UsbStartDevice: bMaxPacketSize0 0x%x\n", deviceDescriptor->bMaxPacketSize0));
                KdPrint((debugPrefix "UsbStartDevice: idVendor 0x%x\n", deviceDescriptor->idVendor));
                KdPrint((debugPrefix "UsbStartDevice: idProduct 0x%x\n", deviceDescriptor->idProduct));
                KdPrint((debugPrefix "UsbStartDevice: bcdDevice 0x%x\n", deviceDescriptor->bcdDevice));
                KdPrint((debugPrefix "UsbStartDevice: iManufacturer 0x%x\n", deviceDescriptor->iManufacturer));
                KdPrint((debugPrefix "UsbStartDevice: iProduct 0x%x\n", deviceDescriptor->iProduct));
                KdPrint((debugPrefix "UsbStartDevice: iSerialNumber 0x%x\n", deviceDescriptor->iSerialNumber));
                KdPrint((debugPrefix "UsbStartDevice: bNumConfigurations 0x%x\n", deviceDescriptor->bNumConfigurations));
                KdPrint((debugPrefix "UsbStartDevice: =======================\n"));
            }
        }
        else
            ntStatus = STATUS_NO_MEMORY;

        if (NT_SUCCESS(ntStatus))
        {
            // maak een pointer naar de Device Descriptor in de Device Extension voor gemakkelijke toegang.
            // Dit geheugen zal worden vrijgegeven wanneer het device wordt verwijderd.
            // Zie "UsbRemoveDevice".
            pDeviceExtension->DeviceDescriptor = deviceDescriptor;
            pDeviceExtension->Stopped = FALSE;
        }
        else if (deviceDescriptor)
        {
            // als de bustransactie mislukt is, geef dan het geheugen vrij
            // dat gereserveerd was om de Device Descriptor in te bewaren,
            // omdat het device waarschijnlijk niet functioneert
            ExFreePool(deviceDescriptor);
            pDeviceExtension->DeviceDescriptor = NULL;
        }
        ExFreePool(urb);
    }
    else
        // geheugen verkrijgen voor URB is mislukt
        ntStatus = STATUS_NO_MEMORY;

    // als de GetDescriptor aanroep succesvol was, herconfigureer dan het device
    if (NT_SUCCESS(ntStatus))
        ntStatus = UsbConfigureDevice(DeviceObject);

    KdPrint((debugPrefix "UsbStartDevice: einde (%x)\n", ntStatus));
    return ntStatus;
}

NTSTATUS
UsbStopDevice(
   IN  PDEVICE_OBJECT DeviceObject
   )
/*
Routine Description:
   Stopt een bepaalde instantie van het USB device op de USB bus.
Arguments:
   DeviceObject - pointer naar het Device Object voor deze instantie van het USB device
Return Value:
   NT status code
*/
{
   NTSTATUS ntStatus = STATUS_SUCCESS;
   PDEVICE_EXTENSION pDeviceExtension = (PDEVICE_EXTENSION)DeviceObject->DeviceExtension;
   ULONG urbSize = sizeof(struct _URB_SELECT_CONFIGURATION);
   PURB urb;
   KdPrint((debugPrefix "UsbStopDevice: begin\n"));

   if (urb = ExAllocatePool(NonPagedPool, urbSize))
   {
      // verstuur een "Select Configuration" URB met een NULL pointer als configuratiehandle.
      // zo wordt de configuratie gesloten en het device in de "ongeconfigureerde" toestand
      // geplaatst
      UsbBuildSelectConfigurationRequest(urb, (USHORT)urbSize, NULL);
      KdPrint((
         debugPrefix "UsbStopDevice: Device Configuration Closed, status = %x, USB status = %x\n",
         UsbCallUSBD(DeviceObject, urb),
         urb->UrbHeader.Status
      ));
      ExFreePool(urb);
   }
   else
      ntStatus = STATUS_NO_MEMORY;

   KdPrint((debugPrefix "UsbStopDevice: einde (ntStatus = %x)\n", ntStatus));
   return ntStatus;
}

// de volgende functie is niet relevant en dient uitsluitend voor debugging doeleinden
NTSTATUS
UsbResetParentPort(
    IN PDEVICE_OBJECT DeviceObject
    )
/*
Routine Description:
    Reset de parentpoort
Arguments:

Return Value:
    STATUS_SUCCESS indien succesvol,
    anders STATUS_UNSUCCESSFUL
*/
{
    NTSTATUS ntStatus, status = STATUS_SUCCESS;
    PIRP irp;
    KEVENT event;
    IO_STATUS_BLOCK ioStatus;
    PIO_STACK_LOCATION nextStack;
    PDEVICE_EXTENSION pDeviceExtension = (PDEVICE_EXTENSION)DeviceObject->DeviceExtension;
    KdPrint((debugPrefix "UsbResetParentPort: begin\n"));

    // genereer een synchroon verzoek
    KeInitializeEvent(&event, NotificationEvent, FALSE);

    irp = IoBuildDeviceIoControlRequest(
             IOCTL_INTERNAL_USB_RESET_PORT,
             pDeviceExtension->StackDeviceObject,
//             pDeviceExtension->TopOfStackDeviceObject,
             NULL,
             0,
             NULL,
             0,
             TRUE, /* INTERNAL */
             &event,
             &ioStatus);

    // roep de klassedriver aan om de bewerking uit te voeren.
    // als de teruggegeven status "PENDING" is, wacht dan
    // totdat het verzoek compleet is.
    nextStack = IoGetNextIrpStackLocation(irp);
    ASSERT(nextStack != NULL);

    KdPrint((debugPrefix "UsbResetParentPort: USBD Enable Port API wordt aangeroepen\n"));
    ntStatus = IoCallDriver(pDeviceExtension->StackDeviceObject, irp);
    KdPrint((debugPrefix "UsbResetParentPort: USBD Enable Port API is aangeroepen (ntStatus = %x)\n", ntStatus));

    if (ntStatus == STATUS_PENDING)
    {
        KdPrint((debugPrefix "UsbResetParentPort: KeWaitForSingleObject wordt aangeroepen\n"));
        status = KeWaitForSingleObject(
                       &event,
                       Suspended,
                       KernelMode,
                       FALSE,
                       NULL);
        KdPrint((debugPrefix "UsbResetParentPort: KeWaitForSingleObject is aangeroepen (ntStatus = %x)\n", status));
    }
    else
        ioStatus.Status = ntStatus;

    // USBD mapt de foutcode
    ntStatus = ioStatus.Status;

    KdPrint((debugPrefix "UsbResetParentPort: einde (ntStatus = %x)\n", ntStatus));
    return ntStatus;
}

// de volgende functie is niet relevant en dient uitsluitend voor debugging doeleinden
ULONG
UsbGetCurrentFrameNumber(
    IN PDEVICE_OBJECT DeviceObject
    )
{
   PURB              urb;
   PDEVICE_EXTENSION pDeviceExtension = (PDEVICE_EXTENSION)DeviceObject->DeviceExtension;
   NTSTATUS          ntStatus = STATUS_SUCCESS;
   ULONG             frameNumber = 0;
   KdPrint((debugPrefix "UsbGetCurrentFrameNumber: begin\n"));

   if (urb = ExAllocatePool(NonPagedPool, sizeof(struct _URB_GET_CURRENT_FRAME_NUMBER)))
   {
      urb->UrbHeader.Function = URB_FUNCTION_GET_CURRENT_FRAME_NUMBER;
      urb->UrbHeader.Length = sizeof(struct _URB_GET_CURRENT_FRAME_NUMBER);
      if (NT_SUCCESS(ntStatus = UsbCallUSBD(DeviceObject, urb)))
         frameNumber = urb->UrbGetCurrentFrameNumber.FrameNumber;
      ExFreePool(urb);
   }

   KdPrint((debugPrefix "UsbGetCurrentFrameNumber: einde (resultaat = %d)\n", frameNumber));
   return frameNumber;
} // UsbGetCurrentFrameNumber

NTSTATUS
UsbCallUSBD(
    IN PDEVICE_OBJECT DeviceObject,
    IN PURB Urb
    )
/*
Routine Description:
   Geeft een Usb Request Block (URB) door aan de USB klassedriver (USBD)

   Merk op dat hier een nieuwe IRP gemaakt wordt en gebruikt wordt om het
   verzoek te versturen naar het USB software systeem. Dit betekent dat deze
   routine in essentie onafhankelijk is van de IRP die aan het begin
   veroorzaakte dat deze driver werd aangeroepen.De IRP voor deze overdracht
   wordt gemaakt, gebruik en dan vernietigd binnen deze routine.

   Maar deze routine gebruikt wel het USB Request Block (URB) afkomstig
   van de aanroeper als het Request Block voor de USB software stack.

   De implementatie van deze routine kan veranderd worden afhankelijk van
   de sepcifieke drivereisen. Deze routine geeft bijvoorbeeld synchrone
   verzoeken aan de USB stack, terwijl dit misschien als asynchroon verzoek
   gemplementeerd moet worden waarbij een IoCompletionRoutine ingesteld wordt
   die aangeroepen wordt wanneer het verzoek compleet is.
Arguments:
   DeviceObject - pointer naar het Device Object voor deze instantie van het USB device
   Urb          - pointer naar USB Request Block
Return Value:
   STATUS_SUCCESS indien succesvol,
   anders STATUS_UNSUCCESSFUL
*/
{
   NTSTATUS ntStatus, status = STATUS_SUCCESS;
   PDEVICE_EXTENSION pDeviceExtension = (PDEVICE_EXTENSION)DeviceObject->DeviceExtension;
   PIRP irp;
   KEVENT event;
   IO_STATUS_BLOCK ioStatus;
   PIO_STACK_LOCATION nextStack;
   KdPrint((debugPrefix "UsbCallUSBD: begin\n"));

   // geef een synchroon verzoek (zie opmerkingen hierboven)
   KeInitializeEvent(&event, NotificationEvent, FALSE);

   irp = IoBuildDeviceIoControlRequest(
             IOCTL_INTERNAL_USB_SUBMIT_URB,
             pDeviceExtension->StackDeviceObject,
             NULL,
             0,
             NULL,
             0,
             TRUE, // INTERNAL
             &event,
             &ioStatus);

   // voorbereiden om de USB driver stack aan te roepen
   nextStack = IoGetNextIrpStackLocation(irp);
   ASSERT(nextStack != NULL);

   // maak de URB pointer klaar om door te geven aan de USB driver stack
   nextStack->Parameters.Others.Argument1 = Urb;

   KdPrint((debugPrefix "UsbCallUSBD: IoCallDriver wordt aangeroepen voor het benaderen van de USB driver stack\n"));

   // roep de USB class driver aan om de bewerking uit te voeren.
   // indien de teruggegeven status PENDING is, wacht dan tot het verzoek compleet is.
   ntStatus = IoCallDriver(pDeviceExtension->StackDeviceObject, irp);

   KdPrint((debugPrefix "UsbCallUSBD: IoCallDriver is aangeroepen (ntStatus = %x)\n", ntStatus));

   if (ntStatus == STATUS_PENDING)
   {
      KdPrint((debugPrefix "UsbCallUSBD: KeWaitForSingleObject wordt aangeroepen\n"));

      status = KeWaitForSingleObject(
                    &event,
                    Suspended,
                    KernelMode,
                    FALSE,
                    NULL);

      KdPrint((debugPrefix "UsbCallUSBD: KeWaitForSingleObject is aangeroepen (resultaat = %x)\n", status));
   }
   else
      ioStatus.Status = ntStatus;

   KdPrint((debugPrefix "UsbCallUSBD: URB status = %x\n", Urb->UrbHeader.Status));

   // USBD mapt de foutcode. USBD gebruikt foutcodes in de URB structuur
   // die duidelijker samenhangen met USB gedrag. Om de NT statuscodes te laten
   // overeenkomen mapt USBD de foutcodes naar algemenere NT foutcategorien
   // zodat hogerliggende drivers de foutcodes kunnen ontcijferen op basis van
   // standaard NT foutcode definities.
   ntStatus = ioStatus.Status;

   // indien de URB status niet USBD_STATUS_SUCCESS is, slaan we een kopie van de
   // URB status op in de Device Extension. Na een fout kan een andere IOCTL
   // IOCTL_EZUSB_GET_LAST_ERROR gebruikt worden om de URB status van de meest recent
   // gefaalde URB opgevraagd worden. Natuurlijk wordt deze status overschreven door
   // volgende fouten, maar het is beter dan niets.
   if (!(USBD_SUCCESS(Urb->UrbHeader.Status)))
      pDeviceExtension->LastFailedUrbStatus = Urb->UrbHeader.Status;

   // als ioStatus.Status een fout aangeeft (bijvoorbeeld als de IRP mislukt is)
   // dan wordt die teruggegeven.
   // als ioStatus.Status succes aangeeft, dan is het nog steeds mogelijk dat onze
   // actie mislukt is. bijvoorbeeld als de IRP geannuleerd is, geeft de status
   // teruggegeven door de I/O manager geen fout aan. in dat geval moet de URB
   // status gecontroleerd worden. als het iets anders aangeeft dan USBD_SUCCESS,
   // moet STATUS_UNSUCCESSFUL teruggegeven worden.
   if (NT_SUCCESS(ntStatus))
      if (!(USBD_SUCCESS(Urb->UrbHeader.Status)))
         ntStatus = STATUS_UNSUCCESSFUL;

   KdPrint((debugPrefix "UsbCallUSBD: einde (ntStatus = %x)\n", ntStatus));
   return ntStatus;
}
