#include "driver.h"
#include "usb.h"
#include "usbezusb.h"

NTSTATUS
Ezusb8051Reset(
   PDEVICE_OBJECT DeviceObject,
   UCHAR resetBit
   )
/*
Routine Description:
   Gebruikt het ANCHOR LOAD vendorspecifieke commando om het 8051 reset bit
   in de EZ-USB chip te zetten dan wel te wissen.
Arguments:
   DeviceObject - pointer naar het Device Object voor deze instantie van het USB device
   resetBit - 1 zet het 8051 reset bit (houdt de 8051 in reset)
              0 wist het 8051 reset bit (8051 begint te lopen)
Return Value:
   STATUS_SUCCESS indien successful,
   anders STATUS_UNSUCCESSFUL
*/
{
   NTSTATUS ntStatus;
   PURB urb;

   if (urb = ExAllocatePool(NonPagedPool, sizeof(struct _URB_CONTROL_VENDOR_OR_CLASS_REQUEST)))
   {
      // wijzig het EZ-USB reset bit
      urb->UrbHeader.Length = sizeof(struct _URB_CONTROL_VENDOR_OR_CLASS_REQUEST);
      urb->UrbHeader.Function = URB_FUNCTION_VENDOR_DEVICE;

      urb->UrbControlVendorClassRequest.TransferFlags = 0;
      urb->UrbControlVendorClassRequest.TransferBufferLength = 1;
      urb->UrbControlVendorClassRequest.TransferBuffer = &resetBit;
      urb->UrbControlVendorClassRequest.TransferBufferMDL = NULL;
      urb->UrbControlVendorClassRequest.UrbLink = NULL;
      urb->UrbControlVendorClassRequest.RequestTypeReservedBits = 0;
      urb->UrbControlVendorClassRequest.Request = ANCHOR_LOAD_INTERNAL;
      urb->UrbControlVendorClassRequest.Value = CPUCS_REG_EZUSB;
      urb->UrbControlVendorClassRequest.Index = 0;

      ntStatus = UsbCallUSBD(DeviceObject, urb);

      ExFreePool(urb);
   }
   else
      ntStatus = STATUS_NO_MEMORY;

   return ntStatus;
}

// dit is het aantal firmwarebytes dat per "setup" overdracht wordt gedownload
#define CHUNK_SIZE 64

NTSTATUS
EzusbAnchorDownload(
   PDEVICE_OBJECT DeviceObject,
   WORD offset,
   PUCHAR downloadBuffer,
   ULONG downloadSize
   )
/*
Routine Description:
   Uses the ANCHOR LOAD vendor specific command to download code to the EZ-USB
   device. The actual code is stored as data within the driver binary in the
   global 'firmware' which is an EZUSB_FIRMWARE struct included in the file
   firmware.c.
Arguments:
   DeviceObject - pointer naar het Device Object voor deze instantie van het USB device
   downloadBuffer - pointer to the firmware image
   downloadSize - total size (bytes) of the firmware image to download
Return Value:
   STATUS_SUCCESS indien succesvol,
   anders STATUS_UNSUCCESSFUL
*/
{
   NTSTATUS ntStatus;
   PURB urb;
   int i;
   int chunkCount;
   PUCHAR ptr = downloadBuffer;

   if (urb = ExAllocatePool(NonPagedPool, sizeof(struct _URB_CONTROL_VENDOR_OR_CLASS_REQUEST)))
   {
      chunkCount = ((downloadSize + CHUNK_SIZE - 1) / CHUNK_SIZE);
      // The download will be split into CHUNK_SIZE pieces and
      // downloaded with multiple setup transfers.  For the Rev B parts
      // CHUNK_SIZE should not exceed 64 bytes, as larger transfers can
      // result in data corruption when other USB devices are present.
      for (i = 0; i < chunkCount; i++)
      {
         urb->UrbHeader.Length = sizeof(struct _URB_CONTROL_VENDOR_OR_CLASS_REQUEST);
         urb->UrbHeader.Function = URB_FUNCTION_VENDOR_DEVICE;

         urb->UrbControlVendorClassRequest.TransferBufferLength =
            ((i == (chunkCount - 1)) &&  (downloadSize % CHUNK_SIZE))
            ?
            (downloadSize % CHUNK_SIZE)
            :
            CHUNK_SIZE;

         urb->UrbControlVendorClassRequest.TransferFlags = 0;
         urb->UrbControlVendorClassRequest.TransferBuffer = ptr;
         urb->UrbControlVendorClassRequest.TransferBufferMDL = NULL;
         urb->UrbControlVendorClassRequest.UrbLink = NULL;
         urb->UrbControlVendorClassRequest.RequestTypeReservedBits = 0;
         urb->UrbControlVendorClassRequest.Request = ANCHOR_LOAD_INTERNAL;
         urb->UrbControlVendorClassRequest.Value = (i * CHUNK_SIZE) + offset;
         urb->UrbControlVendorClassRequest.Index = 0;

         if (!NT_SUCCESS(ntStatus = UsbCallUSBD(DeviceObject, urb)))
            break;

         ptr += CHUNK_SIZE;
      }
      ExFreePool(urb);
   }
   else
      ntStatus = STATUS_NO_MEMORY;

   return ntStatus;
}
