#include "driver.h"
#include "usb.h"
#include "usbgetdescriptor.h"

ULONG
UsbGetDescriptorWithIndexAndLanguage(
    IN PDEVICE_OBJECT DeviceObject,
    PVOID             pOutputBuffer,
    ULONG             ulOutputBufferLength,
    UCHAR             DescriptorType,
    UCHAR             Index,
    USHORT            LanguageId
    )
/*
    parameters:
       DeviceObject   - pointer naar het Device Object
       pOutputBuffer  - pointer naar de buffer waarin de data gezet moet worden
       ulBufferLength - grootte van de gealloceerde Descriptor buffer,
                        bijvoorbeeld de grootte van structures uit ntddk\inc\usb100.h, zoals:
                        sizeof(USB_DEVICE_DESCRIPTOR)
                        sizeof(USB_CONFIGURATION_DESCRIPTOR)
                        sizeof(USB_STRING_DESCRIPTOR)
                        sizeof(USB_INTERFACE_DESCRIPTOR)
                        sizeof(USB_ENDPOINT_DESCRIPTOR)
                        sizeof(USB_CONFIGURATION_POWER_DESCRIPTOR)
                        sizeof(USB_INTERFACE_POWER_DESCRIPTOR)
                        sizeof(USB_COMMON_DESCRIPTOR)
                        sizeof(USB_HUB_DESCRIPTOR)
       DescriptorType - Descriptor Type volgende de USB specificatie,
                        bijvoorbeeld constanten uit ntddk\inc\usb100.h, zoals:
                        USB_DEVICE_DESCRIPTOR_TYPE
                        USB_CONFIGURATION_DESCRIPTOR_TYPE
                        USB_STRING_DESCRIPTOR_TYPE
                        USB_INTERFACE_DESCRIPTOR_TYPE
                        USB_ENDPOINT_DESCRIPTOR_TYPE
                        USB_RESERVED_DESCRIPTOR_TYPE
                        USB_CONFIG_POWER_DESCRIPTOR_TYPE
                        USB_INTERFACE_POWER_DESCRIPTOR_TYPE
       Index          - index van de String Descriptor
       LanguageId     - Language ID van de String Descriptor
   post:
      indien succesvol dan UsbGetDescriptor = het aantal geldige bytes in de data buffer
      anders UsbGetDescriptor = 0;
   effect:
      leest de aangegeven Descriptor van het opgegeven Device Object en
      maakt daarbij gebruik van de opgegeven index en Language ID
*/
{
   NTSTATUS ntStatus = STATUS_SUCCESS;
   PURB     urb      = NULL;
   ULONG    length   = 0;
   KdPrint((debugPrefix "UsbGetDescriptorWithIndexAndLanguage: begin (Index = %d, LanguageId = %d)\n", Index, LanguageId));

   if (urb = ExAllocatePool(NonPagedPool, sizeof(struct _URB_CONTROL_DESCRIPTOR_REQUEST)))
   {
      if (pOutputBuffer)
      {

         UsbBuildGetDescriptorRequest(urb,
                                      (USHORT)sizeof(struct _URB_CONTROL_DESCRIPTOR_REQUEST),
                                      DescriptorType,       // Descriptor type
                                      Index,                // index
                                      LanguageId,           // Language ID
                                      pOutputBuffer,        // transferbuffer
                                      NULL,                 // MDL
                                      ulOutputBufferLength, // bufferlengte
                                      NULL);                // link
         // doe de aanroep en indien succesvol, geef dan de lengte van de URB terug
         if (NT_SUCCESS(ntStatus = UsbCallUSBD(DeviceObject, urb)))
            length = urb->UrbControlDescriptorRequest.TransferBufferLength;
      }
      else
         ntStatus = STATUS_INVALID_PARAMETER;

      ExFreePool(urb);
   }
   else
      ntStatus = STATUS_NO_MEMORY;        
    
   KdPrint((debugPrefix "UsbGetDescriptorWithIndexAndLanguage: einde (resultaatlengte = %d, ntStatus = %d)\n", length, ntStatus));
   return length;
}

PUSB_CONFIGURATION_DESCRIPTOR
UsbGetConfigDescriptor(
    IN PDEVICE_OBJECT DeviceObject
    )
/*
Routine Description:

Arguments:

Return Value:
    NT status code
*/
{
    NTSTATUS ntStatus;
    PURB urb;
    ULONG size;
    PUSB_CONFIGURATION_DESCRIPTOR configurationDescriptor = NULL;
    KdPrint((debugPrefix "UsbGetConfigDescriptor: begin\n"));

    if (urb = ExAllocatePool(NonPagedPool, sizeof(struct _URB_CONTROL_DESCRIPTOR_REQUEST)))
    {
        size = sizeof(USB_CONFIGURATION_DESCRIPTOR) + 256;

get_config_descriptor_retry2:
        if (configurationDescriptor = ExAllocatePool(NonPagedPool, size))
        {
            UsbBuildGetDescriptorRequest(urb,
                                         (USHORT)sizeof(struct _URB_CONTROL_DESCRIPTOR_REQUEST),
                                         USB_CONFIGURATION_DESCRIPTOR_TYPE,
                                         0,
                                         0,
                                         configurationDescriptor,
                                         NULL,
                                         size,
                                         NULL);

            ntStatus = UsbCallUSBD(DeviceObject, urb);

            KdPrint((debugPrefix "Configuration Descriptor = %x, len %x\n",
                     configurationDescriptor,
                     urb->UrbControlDescriptorRequest.TransferBufferLength));
        }
        else
            ntStatus = STATUS_INSUFFICIENT_RESOURCES;

        if (
              NT_SUCCESS(ntStatus)
              &&
              (urb->UrbControlDescriptorRequest.TransferBufferLength >= sizeof(USB_CONFIGURATION_DESCRIPTOR))
              &&
              (configurationDescriptor->wTotalLength >= sizeof(USB_CONFIGURATION_DESCRIPTOR))
           )
        {
            // The Get Config Descriptor request did not return an error
            // AND at least enough data was transferred to fill a Config
            // Descriptor AND the Config Descriptor wLength is at least the
            // size of a Config Descriptor
            if (configurationDescriptor->wTotalLength > size)
            {
                // The request buffer is not big enough to hold the
                // entire set of descriptors.  Free the current buffer
                // and retry with a buffer which should be big enough.
                size = configurationDescriptor->wTotalLength;
                ExFreePool(configurationDescriptor);
                configurationDescriptor = NULL;
                goto get_config_descriptor_retry2;
            }
            else if (configurationDescriptor->wTotalLength
                     >
                     urb->UrbControlDescriptorRequest.TransferBufferLength)
            {
                // The request buffer is greater than or equal to the
                // Config Descriptor wLength, but less data was transferred
                // than wLength.  Return NULL to indicate a device error.
                ExFreePool(configurationDescriptor);
                configurationDescriptor = NULL;
            }
            // else  everything is OK with the Config Descriptor, return it.
        }
        else
        {
            // The Get Config Descriptor request returned an error OR
            // not enough data was transferred to fill a Config Descriptor
            // OR the Config Descriptor wLength is less than the size of
            // a Config Descriptor.  Return NULL to indicate a device error.
            ExFreePool(configurationDescriptor);
            configurationDescriptor = NULL;
        }

        ExFreePool(urb);
    }
    else
        ntStatus = STATUS_INSUFFICIENT_RESOURCES;

    KdPrint((debugPrefix "UsbGetConfigDescriptor: einde\n"));
    return configurationDescriptor;
}
