#include "driver.h"
#include "utility.h"
#include "ioctl.h"
#include "usb.h"
#include "usbgetdescriptor.h"
#include "usbrequests.h"

NTSTATUS
UsbSelectInterfaces(
    IN PDEVICE_OBJECT DeviceObject,
    IN PUSB_CONFIGURATION_DESCRIPTOR ConfigurationDescriptor,
    IN PUSBD_INTERFACE_INFORMATION Interface
    )
/*
Routine Description:
    Initialiseert een USB Device met meerdere interfaces
Arguments:
    DeviceObject            - pointer naar het Device Object voor deze instantie van het USB Device
    ConfigurationDescriptor - pointer naar de USB Configuration Descriptor die de Interface en Endpoint Descriptors bevat.
    Interface               - pointer naar een USBD Interface Information Object
                            - indien dit NULL is, dan moet de driver de interface kiezen op basis van driverspecifieke criteria
                              en moet de driver het device ook CONFIGUREREN.
                            - indien dit NIET NULL is, dan is aan de driver al een interface toegewezen en
                              is het apparaat al geconfigureerd door de parent van deze device driver.
Return Value:
    NT status code
*/
{
   PDEVICE_EXTENSION pDeviceExtension = (PDEVICE_EXTENSION)DeviceObject->DeviceExtension;
   NTSTATUS ntStatus;
   PURB urb;
   ULONG j;
   UCHAR alternateSetting;
   UCHAR MyInterfaceNumber = SAMPLE_INTERFACE_NBR;
   PUSBD_INTERFACE_INFORMATION interfaceObject;
   USBD_INTERFACE_LIST_ENTRY interfaceList[2];
   KdPrint((debugPrefix "UsbSelectInterfaces: begin\n"));

   // zoek in de configuratiedescriptor naar de eerste Interface/Alternate Setting
   interfaceList[0].InterfaceDescriptor =
   USBD_ParseConfigurationDescriptorEx(ConfigurationDescriptor,
                                       ConfigurationDescriptor,
                                       -1,         // Interface - niet relevant
                                       -1,         // Alternate Setting - niet relevant
                                       -1,         // Class - niet relevant
                                       -1,         // SubClass - niet relevant
                                       -1);        // Protocol - niet relevant

   ASSERT(interfaceList[0].InterfaceDescriptor != NULL);

   interfaceList[1].InterfaceDescriptor = NULL;
   interfaceList[1].Interface = NULL;

   if (!(urb = USBD_CreateConfigurationRequestEx(ConfigurationDescriptor, &interfaceList[0])))
      KdPrint((debugPrefix "UsbSelectInterfaces: USBD_CreateConfigurationRequestEx mislukt\n"));

   KdPrint((debugPrefix "UsbSelectInterfaces: inhoud URB (voor):\n"));
   UtilDumpBuffer(urb, urb->UrbHeader.Length);

   interfaceObject = (PUSBD_INTERFACE_INFORMATION)(&(urb->UrbSelectConfiguration.Interface));

   // er moet een standaard maximale overdrachtsgrootte ingesteld worden voor de endpoints.
   // in deze driver moet dit worden aangepast zodat dit de mogelijkheden van de endpoints
   // van het apparaat reflecteert.
   for (j = 0; j < interfaceList[0].InterfaceDescriptor->bNumEndpoints; j++)
      interfaceObject->Pipes[j].MaximumTransferSize = (64 * 1024) - 1;

   ntStatus = UsbCallUSBD(DeviceObject, urb);

   KdPrint((debugPrefix "UsbSelectInterfaces: inhoud URB (na):\n"));
   UtilDumpBuffer(urb, urb->UrbHeader.Length);

   if (NT_SUCCESS(ntStatus) && USBD_SUCCESS(urb->UrbHeader.Status))
   {
      // sla de ConfigurationHandle voor dit device op
      pDeviceExtension->ConfigurationHandle = urb->UrbSelectConfiguration.ConfigurationHandle;

      pDeviceExtension->Interface = ExAllocatePool(NonPagedPool, interfaceObject->Length);

      // maak een kopie van de interfaceObject-informatie die was teruggegeven
      RtlCopyMemory(pDeviceExtension->Interface, interfaceObject, interfaceObject->Length);

      // print interfaceObject naar de debugger
      KdPrint((debugPrefix "UsbSelectInterfaces: inhoud interfaceObject:\n"));
      KdPrint((debugPrefix "UsbSelectInterfaces: ========================\n"));
      KdPrint((debugPrefix "UsbSelectInterfaces: NumberOfPipes 0x%x\n", pDeviceExtension->Interface->NumberOfPipes));
      KdPrint((debugPrefix "UsbSelectInterfaces: Length 0x%x\n", pDeviceExtension->Interface->Length));
      KdPrint((debugPrefix "UsbSelectInterfaces: Alt Setting 0x%x\n", pDeviceExtension->Interface->AlternateSetting));
      KdPrint((debugPrefix "UsbSelectInterfaces: Interface Number 0x%x\n", pDeviceExtension->Interface->InterfaceNumber));

      // print de pipeinformatie
      for (j = 0; j < interfaceObject->NumberOfPipes; j++)
      {
         PUSBD_PIPE_INFORMATION pipeInformation = &pDeviceExtension->Interface->Pipes[j];

         KdPrint((debugPrefix "UsbSelectInterfaces: ------------------------\n"));
         KdPrint((debugPrefix "UsbSelectInterfaces: PipeType 0x%x\n", pipeInformation->PipeType));
         KdPrint((debugPrefix "UsbSelectInterfaces: EndpointAddress 0x%x\n", pipeInformation->EndpointAddress));
         KdPrint((debugPrefix "UsbSelectInterfaces: MaxPacketSize 0x%x\n", pipeInformation->MaximumPacketSize));
         KdPrint((debugPrefix "UsbSelectInterfaces: Interval 0x%x\n", pipeInformation->Interval));
         KdPrint((debugPrefix "UsbSelectInterfaces: Handle 0x%x\n", pipeInformation->PipeHandle));
         KdPrint((debugPrefix "UsbSelectInterfaces: MaximumTransferSize 0x%x\n", pipeInformation->MaximumTransferSize));
      }
      KdPrint((debugPrefix "UsbSelectInterfaces: ========================\n"));
   }

   KdPrint((debugPrefix "UsbSelectInterfaces: einde (%x)\n", ntStatus));
   return ntStatus;
} // UsbSelectInterfaces

NTSTATUS
UsbConfigureDevice(
    IN  PDEVICE_OBJECT DeviceObject
    )
/*
Routine Description:
   Configureert het USB device via USB-specifieke apparaatverzoeken en
   heeft interactie met het USB software subsysteem.
Arguments:
   DeviceObject - pointer naar het Device Object voor deze instantie van het USB Device
Return Value:
   NT status code
*/
{
   PDEVICE_EXTENSION pDeviceExtension = DeviceObject->DeviceExtension;
   NTSTATUS ntStatus;
   PURB urb;
   ULONG size;
   PUSB_CONFIGURATION_DESCRIPTOR configurationDescriptor = NULL;
   KdPrint((debugPrefix "UsbConfigureDevice: begin\n"));

   // reserveer geheugen voor het USB Request Block (URB)
   if (urb = ExAllocatePool(NonPagedPool, sizeof(struct _URB_CONTROL_DESCRIPTOR_REQUEST)))
   {
      // bepaal de grootte voor de data buffer. Merk op dat er extra ruimte wordt toegevoegd
      // ter bescherming tegen hardwarefouten die ervoor zouden kunnen zorgen dat het device
      // voorbij het einde van de databuffer zou gaan
        size = sizeof(USB_CONFIGURATION_DESCRIPTOR) + 16;

        // reserveer geheugen uit het niet-gepagineerde poolgeheugen voor de databuffer
        if (configurationDescriptor = ExAllocatePool(NonPagedPool, size))
        {
            UsbBuildGetDescriptorRequest(urb,
                                         (USHORT)sizeof(struct _URB_CONTROL_DESCRIPTOR_REQUEST),
                                         USB_CONFIGURATION_DESCRIPTOR_TYPE,
                                         0,
                                         0,
                                         configurationDescriptor,
                                         NULL,
                                         sizeof(USB_CONFIGURATION_DESCRIPTOR), // lees alleen de Configuration Descriptor
                                         NULL);

            ntStatus = UsbCallUSBD(DeviceObject, urb);

            if (NT_SUCCESS(ntStatus))
            {
                KdPrint((debugPrefix "UsbConfigureDevice: Configuration Descriptor bevindt zich op %x\n",
                         configurationDescriptor));
                KdPrint((debugPrefix "UsbConfigureDevice: bytes verzonden: %d, feitelijke lengte: %d\n",
                         urb->UrbControlDescriptorRequest.TransferBufferLength,
                         configurationDescriptor->wTotalLength));
            }
        }
        else
        {
            ntStatus = STATUS_NO_MEMORY;
            goto UsbConfigureDevice_end;
        }

        // bepaal hoeveel data de gehele Configuration Descriptor bevat
        // en voeg extra ruimte toe ter beveiliging tegen onbedoelde bufferoverloop
        size = configurationDescriptor->wTotalLength + 16;

        // maak het databuffergeheugen vrij dat net gebruikt is
        ExFreePool(configurationDescriptor);
        configurationDescriptor = NULL;

        // reserveer geheugen uit de niet-gepagineerde pool voor de data buffer
        configurationDescriptor = ExAllocatePool(NonPagedPool, size);

        // lees nu de gehele Configuration Descriptor in
        if (configurationDescriptor != NULL)
        {
            UsbBuildGetDescriptorRequest(urb,
                                         (USHORT)sizeof(struct _URB_CONTROL_DESCRIPTOR_REQUEST),
                                         USB_CONFIGURATION_DESCRIPTOR_TYPE,
                                         0,
                                         0,
                                         configurationDescriptor,
                                         NULL,
                                         size,  // Get all the descriptor data
                                         NULL);

            ntStatus = UsbCallUSBD(DeviceObject, urb);

            if (NT_SUCCESS(ntStatus))
                KdPrint((debugPrefix "UsbConfigureDevice: gehele Configuration Descriptor bevindt zich op %x, bytes verzonden: %d\n",
                         configurationDescriptor,
                         urb->UrbControlDescriptorRequest.TransferBufferLength));
            else
                // fout bij het opvragen van de Configuration Descriptor
                goto UsbConfigureDevice_end;

        }
        else
        {
            // mislukt om data buffer (Configuration Descriptor) geheugen te verkrijgen
            ntStatus = STATUS_NO_MEMORY;
            goto UsbConfigureDevice_end;
        }

    }
    else
    {
        // mislukt om URB geheugen te verkrijgen
        ntStatus = STATUS_NO_MEMORY;
        goto UsbConfigureDevice_end;
    }

    // we hebben nu de Configuration Descriptor voor de configuratie die we willen.

    // geef nu het SelectConfiguration commando om de pipes te verkrijgen
    // die bij deze configuratie horen.
    if (configurationDescriptor)
        // verkrijg de pipes
        ntStatus = UsbSelectInterfaces(DeviceObject,
                                       configurationDescriptor,
                                       NULL // device is nog niet geconfigureerd
                                       );

UsbConfigureDevice_end:
    // opschonen en de routine verlaten
    if (urb != NULL)
        ExFreePool(urb); // geef het geheugen voor de URB vrij

    if (configurationDescriptor != NULL)
        ExFreePool(configurationDescriptor); // geef het geheugen van de databuffer vrij

    KdPrint((debugPrefix "UsbConfigureDevice: einde (%x)\n", ntStatus));
    return ntStatus;
} // UsbConfigureDevice

NTSTATUS
UsbSetInterface(
   IN PDEVICE_OBJECT DeviceObject,
   IN UCHAR InterfaceNumber,
   IN UCHAR AlternateSetting
   )
{
   PUSBD_INTERFACE_INFORMATION interfaceInformation = NULL;
   PUSB_INTERFACE_DESCRIPTOR interfaceDescriptor = NULL;
   PUSB_CONFIGURATION_DESCRIPTOR configurationDescriptor;
   PDEVICE_EXTENSION pdx = DeviceObject->DeviceExtension;
   PURB urb = NULL;
   USHORT urbSize;
   ULONG numberOfPipes;
   ULONG i,j;
   NTSTATUS ntStatus;

   // lees de Configuration Descriptor
   configurationDescriptor = UsbGetConfigDescriptor(DeviceObject);
   if (!configurationDescriptor)
   {
      KdPrint((debugPrefix "UsbGetConfigDescriptor mislukt\n"));
      ntStatus = STATUS_UNSUCCESSFUL;
      goto CleanupSetInterface;
   }

   // lees de Interface Descriptor voor de juiste interface
   interfaceDescriptor = USBD_ParseConfigurationDescriptorEx(
                           configurationDescriptor,
                           configurationDescriptor,
                           InterfaceNumber,
                           AlternateSetting,
                           -1,
                           -1,
                           -1);

   if (!interfaceDescriptor)
   {
      ntStatus = STATUS_UNSUCCESSFUL;
      goto CleanupSetInterface;
   }

   numberOfPipes = interfaceDescriptor->bNumEndpoints;
   KdPrint((debugPrefix "numberOfPipes = %d\n", numberOfPipes));

   urbSize = GET_SELECT_INTERFACE_REQUEST_SIZE(numberOfPipes);
   KdPrint((debugPrefix "urbSize = %d\n", urbSize));

   if (!(urb = ExAllocatePool(NonPagedPool, urbSize)))
   {
      ntStatus = STATUS_NO_MEMORY;
      goto CleanupSetInterface;
   }
   
   UsbBuildSelectInterfaceRequest(urb,
                                  urbSize,
                                  pdx->ConfigurationHandle,
                                  InterfaceNumber,
                                  AlternateSetting);

   interfaceInformation = &(urb->UrbSelectInterface.Interface);
   interfaceInformation->Length = GET_USBD_INTERFACE_SIZE(numberOfPipes);

   for (i = 0; i < numberOfPipes; i++)
      interfaceInformation->Pipes[i].MaximumTransferSize = 64 * 1024 - 1;

   ntStatus = UsbCallUSBD(DeviceObject, urb);

   // indien succesvol, update dan de Interface structure in
   // de Device Extension.
   if (NT_SUCCESS(ntStatus))
   {
      for (j = 0; j < interfaceInformation->NumberOfPipes; j++)
      {
         PUSBD_PIPE_INFORMATION pipeInformation;

         pipeInformation = &interfaceInformation->Pipes[j];

         KdPrint((debugPrefix "Inhoud pipeInformation\n"));
         KdPrint((debugPrefix "=======================\n"));
         KdPrint((debugPrefix "PipeType 0x%x\n", pipeInformation->PipeType));
         KdPrint((debugPrefix "EndpointAddress 0x%x\n", pipeInformation->EndpointAddress));
         KdPrint((debugPrefix "MaxPacketSize 0x%x\n", pipeInformation->MaximumPacketSize));
         KdPrint((debugPrefix "Interval 0x%x\n", pipeInformation->Interval));
         KdPrint((debugPrefix "Handle 0x%x\n", pipeInformation->PipeHandle));
         KdPrint((debugPrefix "MaximumTransferSize 0x%x\n", pipeInformation->MaximumTransferSize));
         KdPrint((debugPrefix "=======================\n"));
      }

      if (pdx->Interface)
         ExFreePool(pdx->Interface);

      pdx->Interface = NULL;
      pdx->Interface = ExAllocatePool(NonPagedPool, interfaceInformation->Length);
      if (!pdx->Interface)
      {
         ntStatus = STATUS_NO_MEMORY;
         goto CleanupSetInterface;
      }
      RtlCopyMemory(pdx->Interface, interfaceInformation, interfaceInformation->Length);
   }

   CleanupSetInterface:

   // opruimen en de routine verlaten
   if (urb != NULL)
      ExFreePool(urb);

   if (configurationDescriptor != NULL)
      ExFreePool(configurationDescriptor);

   return(ntStatus);
}   

NTSTATUS
UsbSetFeature(
   IN PDEVICE_OBJECT DeviceObject,
   IN PSET_FEATURE_CONTROL setFeatureControl
   )
/*
   Routine Description:
      Deze routine voert een Set Feature "control" overdracht uit
   Arguments:
      DeviceObject - het Device Object
      setFeatureControl - een datastructuur die de parameters voor het Set Feature commando bevat
   Return Value:
      NTSTATUS 
*/
{
   NTSTATUS          ntStatus = STATUS_SUCCESS;
   PURB              urb      = NULL;
   ULONG             length   = 0;
   PDEVICE_EXTENSION pdx      = DeviceObject->DeviceExtension;
   KdPrint((debugPrefix "UsbSetFeature: begin\n"));

   if (urb = ExAllocatePool(NonPagedPool, sizeof(struct _URB_CONTROL_FEATURE_REQUEST)))
   {
      UsbBuildFeatureRequest(urb, URB_FUNCTION_SET_FEATURE_TO_DEVICE, setFeatureControl->FeatureSelector, setFeatureControl->Index, NULL);
                                                                  
      ntStatus = UsbCallUSBD(DeviceObject, urb);
   }
   else
      ntStatus = STATUS_NO_MEMORY;        
 
   KdPrint((debugPrefix "UsbSetFeature: einde\n"));
   return ntStatus;
}

NTSTATUS
UsbResetPipe(
   IN PDEVICE_OBJECT DeviceObject,
   ULONG PipeNum
   )
/*
Routine Description:
   De "stall" conditie van de pipe PipeNum wordt gereset en
   de "data toggle" op de host wordt gereset.
Arguments:

Return Value:

*/
{
   NTSTATUS ntStatus;
   PDEVICE_EXTENSION           pDeviceExtension = DeviceObject->DeviceExtension;
   PUSBD_INTERFACE_INFORMATION interfaceInfo = NULL;
   USBD_PIPE_HANDLE            pipeHandle = NULL;
   PURB urb;
   KdPrint((debugPrefix "UsbResetPipe: begin\n"));

   // controleer of de geselecteerde pipe geldig is en
   // verkrijg er een handle naar. Als er iets fout is, geef dan een fout
   if (!(interfaceInfo = pDeviceExtension->Interface))
   {
      KdPrint((debugPrefix "UsbResetPipe: geen interface-informatie - wordt beindigd\n"));
      return STATUS_UNSUCCESSFUL;
   }
   
   if (PipeNum > interfaceInfo->NumberOfPipes)
   {
      KdPrint((debugPrefix "UsbResetPipe: ongeldige pipe - wordt beindigd\n"));
      return STATUS_INVALID_PARAMETER;
   }

   pipeHandle = interfaceInfo->Pipes[PipeNum].PipeHandle;

   if (urb = ExAllocatePool(NonPagedPool, sizeof(struct _URB_PIPE_REQUEST)))
   {
      urb->UrbHeader.Length = (USHORT)sizeof(struct _URB_PIPE_REQUEST);
      urb->UrbHeader.Function = URB_FUNCTION_RESET_PIPE;
      urb->UrbPipeRequest.PipeHandle = pipeHandle;

      ntStatus = UsbCallUSBD(DeviceObject, urb);

      ExFreePool(urb);
   }
   else
      ntStatus = STATUS_INSUFFICIENT_RESOURCES;

   KdPrint((debugPrefix "UsbResetPipe: einde\n"));
   return ntStatus;
}

NTSTATUS
UsbAbortPipe(
    IN PDEVICE_OBJECT DeviceObject,
    IN USBD_PIPE_HANDLE PipeHandle
    )
/*
Routine Description:
   Annuleert wachtende overdrachten voor een bepaalde pipe.
Arguments:

Return Value:

*/
{
   NTSTATUS ntStatus;
   PURB urb;
   KdPrint((debugPrefix "AbortPipe: begin\n"));

   if (urb = ExAllocatePool(NonPagedPool, sizeof(struct _URB_PIPE_REQUEST)))
   {
      urb->UrbHeader.Length = (USHORT)sizeof(struct _URB_PIPE_REQUEST);
      urb->UrbHeader.Function = URB_FUNCTION_ABORT_PIPE;
      urb->UrbPipeRequest.PipeHandle = PipeHandle;

      ntStatus = UsbCallUSBD(DeviceObject, urb);

      ExFreePool(urb);
   }
   else
      ntStatus = STATUS_INSUFFICIENT_RESOURCES;

   KdPrint((debugPrefix "AbortPipe: einde\n"));
   return ntStatus;
}

ULONG
UsbVendorRequest(
    IN PDEVICE_OBJECT DeviceObject,
    IN PVENDOR_REQUEST_IN pVendorRequest
    )
{
    NTSTATUS ntStatus = STATUS_SUCCESS;
    PURB     urb      = NULL;
    ULONG    length   = 0;
    PUCHAR   buffer   = NULL;
    ULONG    transferFlags = 0;
    PVOID    transferBuffer = NULL;
    KdPrint((debugPrefix "UsbVendorRequest: begin\n"));

    if (urb = ExAllocatePool(NonPagedPool, sizeof(struct _URB_CONTROL_VENDOR_OR_CLASS_REQUEST)))
    {
        // zeer rommelig. het idee is: als het een "IN" overdracht is, dan is er een buffer doorgegeven
        // vanuit user mode. gebruik dan dus de pointer naar de systeeembuffer als de overdrachtbuffer.
        // als de overdracht uitgaand is, dan moet er een transfer buffer gealloceerd worden.
        // als de lengte van de overdracht 1 is, plaats dan pVendorRequest->bData in de buffer.
        // vul anders de buffer met een vermeerderend bytepatroon.
        if (pVendorRequest->direction)
        {
            transferFlags |= USBD_TRANSFER_DIRECTION_IN;
            transferBuffer = pVendorRequest;
        }
        else
        {
            buffer = ExAllocatePool(NonPagedPool, pVendorRequest->wLength);
            transferBuffer = buffer;
            if (pVendorRequest->wLength == 1)
                buffer[0] = pVendorRequest->bData;
            else
            {
                int i;
                PUCHAR ptr = transferBuffer;
                for (i = 0; i < pVendorRequest->wLength; i++)
                {
                    *ptr = (UCHAR)i;
                    ptr++;
                }
            }
        }

        UsbBuildVendorRequest(
           urb,
           URB_FUNCTION_VENDOR_DEVICE,
           sizeof(struct _URB_CONTROL_VENDOR_OR_CLASS_REQUEST),
           transferFlags,
           0,
           pVendorRequest->bRequest,
           pVendorRequest->wValue,
           pVendorRequest->wIndex,
           transferBuffer,
           NULL,
           pVendorRequest->wLength,
           NULL
        );      

        ntStatus = UsbCallUSBD(DeviceObject, urb);

        // geef alleen een lengte terug als dit een IN transactie was
        if (pVendorRequest->direction)
           length = urb->UrbControlVendorClassRequest.TransferBufferLength;

        ExFreePool(urb);
        if (buffer)
            ExFreePool(buffer);
    }

    KdPrint((debugPrefix "UsbVendorRequest: einde\n"));
    return length;
}

NTSTATUS
UsbVendorRequestMDL(
   IN PDEVICE_OBJECT DeviceObject,
   IN PIRP Irp
   )
{
   PDEVICE_EXTENSION          pDeviceExtension = DeviceObject->DeviceExtension;
   NTSTATUS                   ntStatus;
   PVENDOR_OR_CLASS_REQUEST_CONTROL requestControl = (PVENDOR_OR_CLASS_REQUEST_CONTROL)Irp->AssociatedIrp.SystemBuffer;
   PIO_STACK_LOCATION         irpStack = IoGetCurrentIrpStackLocation(Irp);
   ULONG                      bufferLength = irpStack->Parameters.DeviceIoControl.OutputBufferLength;
   PURB                       urb = NULL;
   ULONG                      transferFlags = 0;
   USHORT                     urbFunction = 0;

   // controleer dat de invoerparameter correct is
   // (in ieder geval de juiste grootte heeft)
   if (irpStack->Parameters.DeviceIoControl.InputBufferLength != sizeof(VENDOR_OR_CLASS_REQUEST_CONTROL))
   {
      ntStatus = STATUS_INVALID_PARAMETER;
      return ntStatus;
   }

   // alloceer en vul het USB request (URB)
   if (urb = ExAllocatePool(NonPagedPool, sizeof(struct _URB_CONTROL_VENDOR_OR_CLASS_REQUEST)))
   {
      transferFlags = USBD_SHORT_TRANSFER_OK;

      // verkrijg informatie over de richting uit de invoerparameters
      if (requestControl->direction)
         transferFlags |= USBD_TRANSFER_DIRECTION_IN;

      // het type van het verzoek (Class of Vendor) en de ontvanger
      // (Device, Interface, Endpoint, overige) combine to determine the 
      // URB function.  The following ugly code transforms fields in
      // the input param into an URB function
      switch ((requestControl->requestType << 2) | requestControl->recepient)
      {
         case 0x04:
            urbFunction = URB_FUNCTION_CLASS_DEVICE;
            break;
         case 0x05:
            urbFunction = URB_FUNCTION_CLASS_INTERFACE;
            break;
         case 0x06:
            urbFunction = URB_FUNCTION_CLASS_ENDPOINT;
            break;
         case 0x07:
            urbFunction = URB_FUNCTION_CLASS_OTHER;
            break;
         case 0x08:
            urbFunction = URB_FUNCTION_VENDOR_DEVICE;
            break;
         case 0x09:
            urbFunction = URB_FUNCTION_VENDOR_INTERFACE;
            break;
         case 0x0A:
            urbFunction = URB_FUNCTION_VENDOR_ENDPOINT;
            break;
         case 0x0B:
            urbFunction = URB_FUNCTION_VENDOR_OTHER;
            break;
         default:
            return STATUS_INVALID_PARAMETER;
      }

      UsbBuildVendorRequest(
         urb,
         urbFunction,
         sizeof(struct _URB_CONTROL_VENDOR_OR_CLASS_REQUEST),
         transferFlags,
         0,
         requestControl->request,
         requestControl->value,
         requestControl->index,
         NULL,
         Irp->MdlAddress,
         bufferLength,
         NULL
      );      
   
      // roep de USB stack aan; als de overdracht succesvol is,
      // rapporteer dan de lengte van de overdracht aan de aanroeper
      // door het instellen van IoStatus.Information
      if (NT_SUCCESS(ntStatus = UsbCallUSBD(DeviceObject, urb)))
      {
         Irp->IoStatus.Information = urb->UrbControlVendorClassRequest.TransferBufferLength;
         KdPrint((debugPrefix "overdracht succesvol (0x%x bytes)\n", Irp->IoStatus.Information));
      }

      ExFreePool(urb);

      return ntStatus;
   }
   else
      return STATUS_NO_MEMORY;
}

NTSTATUS
UsbTransfer(
   IN  PDEVICE_OBJECT DeviceObject,
   IN  PIRP Irp
   )
/*
Routine Description:
    
Arguments:

Return Value:
    NT status code
        STATUS_SUCCESS:             lezen was succesvol
        STATUS_INVALID_PARAMETER_3: The Endpoint Index does not specify an IN pipe 
        STATUS_NO_MEMORY:           Insufficient data memory was supplied to perform the READ

    Deze routine vult de statuscode in de IRP
*/
{
   PDEVICE_EXTENSION           pDeviceExtension = DeviceObject->DeviceExtension;
   NTSTATUS                    ntStatus;
   PIO_STACK_LOCATION          irpStack = IoGetCurrentIrpStackLocation(Irp);
   PBULK_TRANSFER_CONTROL      bulkControl = (PBULK_TRANSFER_CONTROL)Irp->AssociatedIrp.SystemBuffer;
   ULONG                       bufferLength = irpStack->Parameters.DeviceIoControl.OutputBufferLength;
   PURB                        urb = NULL;
   ULONG                       urbSize = 0;
   ULONG                       transferFlags = 0;
   PUSBD_INTERFACE_INFORMATION interfaceInfo = pDeviceExtension->Interface;
   PUSBD_PIPE_INFORMATION      pipeInfo = NULL;
   USBD_PIPE_HANDLE            pipeHandle = NULL;
   KdPrint((debugPrefix "UsbTransfer\n"));
   
   // controleer dat de geselecteerde pipe geldig is en verkrijg er een handle voor.
   // als er iets mis is, geef dan een foutmelding.
   if (!interfaceInfo)
   {
      KdPrint((debugPrefix "UsbTransfer: geen interfaceinformatie - functie wordt beindigd\n"));
      return STATUS_UNSUCCESSFUL;
   }
   
   if (bulkControl->pipeNum > interfaceInfo->NumberOfPipes)
   {
      KdPrint((debugPrefix "UsbTransfer: opgegeven pipenummer is te hoog - functie wordt beindigd\n"));
      return STATUS_INVALID_PARAMETER;
   }

   pipeInfo = &(interfaceInfo->Pipes[bulkControl->pipeNum]);
   if (
      !(
         (pipeInfo->PipeType == UsbdPipeTypeBulk)
         ||
         (pipeInfo->PipeType == UsbdPipeTypeInterrupt)
      )
   )
   {
      KdPrint((debugPrefix "UsbTransfer: opgegeven pipe is geen bulk of interrupt pipe - functie wordt beindigd\n"));
      return STATUS_INVALID_PARAMETER;
   }

   if (!(pipeHandle = pipeInfo->PipeHandle))
   {
      KdPrint((debugPrefix "UsbTransfer: ongeldige pipe - functie wordt beindigd\n"));
      return STATUS_UNSUCCESSFUL;
   }

   if (bufferLength > pipeInfo->MaximumTransferSize)
   {
      KdPrint((debugPrefix "UsbTransfer: ongeldige overdrachtsgrootte - functie wordt beindigd\n"));
      return STATUS_INVALID_PARAMETER;
   }

   // alloceer en vul de URB (URB Request)
   if (urb = ExAllocatePool(NonPagedPool, sizeof(struct _URB_BULK_OR_INTERRUPT_TRANSFER)))
   {
      transferFlags = USBD_SHORT_TRANSFER_OK;
      // verkrijg informatie over de richting van het endpointadres
      if (USB_ENDPOINT_DIRECTION_IN(pipeInfo->EndpointAddress))
         transferFlags |= USBD_TRANSFER_DIRECTION_IN;

      UsbBuildInterruptOrBulkTransferRequest(
         urb,                                            // pointer naar URB
         sizeof(struct _URB_BULK_OR_INTERRUPT_TRANSFER), // grootte van URB
	     pipeHandle,                                     // usbd pipe handle
	     NULL,                                           // TransferBuffer
	     Irp->MdlAddress,                                // MDL
         bufferLength,                                   // bufferlengte
	     transferFlags,                                  // flags
	     NULL                                            // link
      );

      // benader de USB stack; als de overdracht succesvol was,
      // geef dan de lengte ervan door aan de aanroeper via IoStatus.Information
      if (NT_SUCCESS(ntStatus = UsbCallUSBD(DeviceObject, urb)))
      {
         Irp->IoStatus.Information = urb->UrbBulkOrInterruptTransfer.TransferBufferLength;
         KdPrint((debugPrefix "UsbTransfer: overdracht succesvol (0x%x bytes)\n",Irp->IoStatus.Information));
      }

      // geef de URB vrij
      ExFreePool(urb);
      return ntStatus;
   }
   else
   {
      KdPrint((debugPrefix "UsbTransfer: kon geen URB alloceren - functie wordt beindigd\n"));
      return STATUS_NO_MEMORY;
   }
}
